#include "dir_sampled.h"

#ifndef pi
#define pi 3.14159265
#endif

#ifndef clogpi
#define clogpi 1.144729885849
#endif

// --------------------------------------------------------------------------
// -- dir_sampled
// --   constructor; initializes to empty
// --------------------------------------------------------------------------
dir_sampled::dir_sampled() :
   D(0), Nthreads(0), alphah(0), N(0), param(0), rArr(NULL),
   temp_counts_itr(NULL), temp_indices_itr(NULL), temp_counts_N(0), temp_counts_invalid(true)
{
   //mexErrMsgTxt("shouldn't be here\n");
}

// --------------------------------------------------------------------------
// -- dir_sampled
// --   copy constructor;
// --------------------------------------------------------------------------
dir_sampled::dir_sampled(const dir_sampled& that)
{
   copy(that);
}
// --------------------------------------------------------------------------
// -- operator=
// --   assignment operator
// --------------------------------------------------------------------------
dir_sampled& dir_sampled::operator=(const dir_sampled& that)
{
   if (this != &that)
   {
      cleanup();
      copy(that);
   }
   return *this;
}
// --------------------------------------------------------------------------
// -- copy
// --   returns a copy of this
// --------------------------------------------------------------------------
void dir_sampled::copy(const dir_sampled& that, bool isClean, bool copyIndices_counts)
{
   if (!isClean)
      cleanup();
   N = that.N;
   Nthreads = that.Nthreads;
   D = that.D;
   alphah = that.alphah;

   if (copyIndices_counts)
      indices_counts = that.indices_counts;

   param = that.param;

   rArr = allocate_memory<gsl_rng*>(Nthreads);
   for (int t=0; t<Nthreads; t++)
      rArr[t] = initialize_gsl_rand(rand());

   temp_counts_itr = NULL;
   temp_indices_itr = NULL;
   temp_counts_N = 0;
   temp_counts_invalid = true;

   gammaln = that.gammaln;
}

// --------------------------------------------------------------------------
// -- dir_sampled
// --   constructor; intializes to all the values given
// --------------------------------------------------------------------------
dir_sampled::dir_sampled(int _Nthreads, int _D, Real _alphah) :
   N(0), Nthreads(_Nthreads), D(_D), alphah(_alphah), param(_D),
   temp_counts_itr(NULL), temp_indices_itr(NULL), temp_counts_N(0), temp_counts_invalid(true)
{
   if (fabs(alphah - (Real)((int)alphah)) < 1e-30)
      gammaln = &gammalnint;
   else
      gammaln = &gsl_sf_lngamma;


   rArr = allocate_memory<gsl_rng*>(Nthreads);
   for (int t=0; t<Nthreads; t++)
      rArr[t] = initialize_gsl_rand(rand());
   clear();
}




// --------------------------------------------------------------------------
// -- ~dir_sampled
// --   destructor
// --------------------------------------------------------------------------
dir_sampled::~dir_sampled()
{
   cleanup();
}
// --------------------------------------------------------------------------
// -- ~cleanup
// --   deletes all the memory allocated by this
// --------------------------------------------------------------------------
void dir_sampled::cleanup()
{
   indices_counts.clear();
   if (rArr)
   {
      for (int t=0; t<Nthreads; t++)
         if (rArr[t]!=NULL)
            gsl_rng_free(rArr[t]);
      deallocate_memory(rArr);
   }

   if (temp_counts_itr!=NULL && temp_counts_N>0)
   {
      delete[] temp_counts_itr;
      delete[] temp_indices_itr;
      temp_counts_itr = NULL;
      temp_indices_itr = NULL;
      temp_counts_N = 0;
      temp_counts_invalid = true;
   }
}

// --------------------------------------------------------------------------
// -- clear
// --   Empties out the statistics and posterior hyperparameters and the
// -- sampled parameters. This basically deletes all notions of data from
// -- the node.
// --------------------------------------------------------------------------
void dir_sampled::clear()
{
   empty();
   update_posteriors();
}
// --------------------------------------------------------------------------
// -- empty
// --   Empties out the statistics and posterior hyperparameters, but not
// -- the sampled parameters
// --------------------------------------------------------------------------
void dir_sampled::empty()
{
   N = 0;
   indices_counts.clear();
}

bool dir_sampled::isempty() const               { return (N==0); }
int dir_sampled::getN() const                   { return N;}
int dir_sampled::getD() const                   { return D;}
arr(Real) dir_sampled::get_logpi() const      { return param.logpi;}
gsl_rng* dir_sampled::get_r()                   { return rArr[0];}

void dir_sampled::set_categorical(categorical &other)
{
   param = other;
}
void dir_sampled::set_categorical(arr(Real) _logpi)
{
   memcpy(param.logpi, _logpi, sizeof(Real)*D);
}

categorical* dir_sampled::get_categorical()
{
   return &param;
}



// --------------------------------------------------------------------------
// -- update_posteriors
// --   Updates the posterior hyperparameters
// --------------------------------------------------------------------------
void dir_sampled::update_posteriors()
{
   temp_counts_invalid = true;
}

void dir_sampled::update_posteriors_sample()
{
   update_posteriors();
   sample();
}

// helper function to merge to indices_counts
int dir_sampled::merge_indices_counts(unordered_map<int,int> &indices_counts2, bool clear)
{
   //for (std::tr1::unordered_map<int,int>::iterator &itr : indices_counts2)
   int numAdded = 0;
   for (unordered_map<int,int>::iterator itr=indices_counts2.begin(); itr!=indices_counts2.end(); itr++)
   {
      int index = itr->first;
      int count = itr->second;
      numAdded += count;

      // try to insert
      std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = indices_counts.emplace(index,count);
      // see if it was found.  if it was found, just add the counts
      if (!itrFnd.second)
         itrFnd.first->second += count;
   }
   if (clear)
      indices_counts2.clear();
   return numAdded;
}

void dir_sampled::temp_merge_with(dir_sampled* other)
{
   if (other!=NULL)
   {
      N += other->N;
      merge_indices_counts(other->indices_counts, false);
      update_posteriors_sample();
      temp_counts_invalid = true;
   }
}

void dir_sampled::merge_with(dir_sampled* other, bool doSample)
{
   if (other!=NULL)
      merge_with(*other, doSample);
   temp_counts_invalid = true;
}
void dir_sampled::merge_with(dir_sampled* other1, dir_sampled* other2, bool doSample)
{
   if (other1==NULL && other2!=NULL)
      merge_with(*other2,doSample);
   else if (other1!=NULL && other2==NULL)
      merge_with(*other1,doSample);
   else if (other1!=NULL && other2!=NULL)
      merge_with(*other1, *other2, doSample);
   temp_counts_invalid = true;
}
void dir_sampled::merge_with(dir_sampled other, bool doSample)
{
   N += other.N;
   merge_indices_counts(other.indices_counts);
   if (doSample)
      update_posteriors_sample();
   else
      update_posteriors();
   temp_counts_invalid = true;
}
void dir_sampled::merge_with(dir_sampled other1, dir_sampled other2, bool doSample)
{
   N += other1.N + other2.N;
   merge_indices_counts(other1.indices_counts);
   merge_indices_counts(other2.indices_counts);
   if (doSample)
      update_posteriors_sample();
   else
      update_posteriors();
   temp_counts_invalid = true;
}
void dir_sampled::set_stats(int _N, const unordered_map<int,int> &_indices_counts)
{
   N = _N;
   indices_counts = _indices_counts;
   //indices_counts.clear();
   //merge_indices_counts(_indices_counts);
   temp_counts_invalid = true;
}
void dir_sampled::set_stats(int _N, const unordered_map<int,int> &indices_counts1, const unordered_map<int,int> &indices_counts2)
{
   N = _N;
   indices_counts = indices_counts1;
   for (unordered_map<int,int>::const_iterator itr=indices_counts2.begin(); itr!=indices_counts2.end(); itr++)
   {
      int index = itr->first;
      int count = itr->second;
      // try to insert
      std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = indices_counts.emplace(index,0);
      itrFnd.first->second += count;
   }
   temp_counts_invalid = true;
}
void dir_sampled::set_stats(int _N, const unordered_map<int,int> &indices_counts1, const unordered_map<int,int> &indices_counts2,
   unordered_map<int,int> &indices_counts3)
{
   N = _N;
   indices_counts = indices_counts1;
   for (unordered_map<int,int>::const_iterator itr=indices_counts2.begin(); itr!=indices_counts2.end(); itr++)
   {
      int index = itr->first;
      int count = itr->second;
      // try to insert
      std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = indices_counts.emplace(index,0);
      itrFnd.first->second += count;
   }
   for (unordered_map<int,int>::const_iterator itr=indices_counts3.begin(); itr!=indices_counts3.end(); itr++)
   {
      int index = itr->first;
      int count = itr->second;
      // try to insert
      std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = indices_counts.emplace(index,0);
      itrFnd.first->second += count;
   }
   temp_counts_invalid = true;
}




Real dir_sampled::data_loglikelihood() const
{
   return param.data_loglikelihood(indices_counts);
}

Real dir_sampled::data_loglikelihood_marginalized()
{
   Real log_likelihood = 0;
   prepare_indices_counts_itr();
   //#pragma omp parallel for reduction(+:log_likelihood)
   for (int i=0; i<temp_counts_N; i++)
   {
      int count = *temp_counts_itr[i];
      log_likelihood += gammaln(count + alphah);
   }

   log_likelihood += -temp_counts_N*gammaln(alphah) + gammaln(D*alphah) - gammaln(N+D*alphah);

   return log_likelihood;
}

Real dir_sampled::data_loglikelihood_marginalized_testmerge(dir_sampled *other) const
{
   Real log_likelihood = 0;
   int nnz = indices_counts.size();
   for (unordered_map<int,int>::const_iterator it=indices_counts.begin(); it!=indices_counts.end(); it++)
   {
      int index = it->first;
      int count = it->second;
      unordered_map<int,int>::const_iterator it2 = other->indices_counts.find(index);
      if (it2!=other->indices_counts.end())
         count += it2->second;
      log_likelihood += gammaln(count + alphah);
   }

   for (unordered_map<int,int>::const_iterator it=other->indices_counts.begin(); it!=other->indices_counts.end(); it++)
   {
      int index = it->first;
      if (indices_counts.count(index)==0)
      {
         nnz++;
         int count = it->second;
         log_likelihood += gammaln(count + alphah);
      }
   }

   log_likelihood += -nnz*gammaln(alphah) + gammaln(D*alphah) - gammaln(N+other->N+D*alphah);
   return log_likelihood;
}


#ifdef USELOGLIKELIHOOD
void dir_sampled::sample()
{
   Real total = 0;
   arr(Real) logpi = param.logpi;
   // first sample all assuming zero counts
   #pragma omp parallel for reduction(+:total)
   for (int d=0; d<D; d++)
   {
      logpi[d] = gsl_ran_gamma(rArr[omp_get_thread_num()], alphah, 1);
      total += logpi[d];
   }
   // now sample the nonzero ones
   prepare_indices_counts_itr();
   #pragma omp parallel for reduction(+:total)
   for (int i=0; i<temp_counts_N; i++)
   {
      int d = temp_indices_itr[i];
      int count = *(temp_counts_itr[i]);
      total -= logpi[d];
      logpi[d] = gsl_ran_gamma(rArr[omp_get_thread_num()], alphah+count, 1);
      total += logpi[d];
   }
   total = log(total);
   #pragma omp parallel for// reduction(+:logmu_prior,logmu_posterior)
   for (int d=0; d<D; d++)
   {
      logpi[d] = log(logpi[d]) - total;
   }
}
#else
void dir_sampled::sample()
{
   Real total = 0;
   arr(Real) logpi = param.logpi;
   // first sample all assuming zero counts
   #pragma omp parallel for reduction(+:total)
   for (int d=0; d<D; d++)
   {
      logpi[d] = gsl_ran_gamma(rArr[omp_get_thread_num()], alphah, 1);
      total += logpi[d];
   }
   // now sample the nonzero ones
   prepare_indices_counts_itr();
   #pragma omp parallel for reduction(+:total)
   for (int i=0; i<temp_counts_N; i++)
   {
      int d = temp_indices_itr[i];
      int count = *(temp_counts_itr[i]);
      total -= logpi[d];
      logpi[d] = gsl_ran_gamma(rArr[omp_get_thread_num()], alphah+count, 1);
      total += logpi[d];
   }
   #pragma omp parallel for// reduction(+:logmu_prior,logmu_posterior)
   for (int d=0; d<D; d++)
   {
      logpi[d] = logpi[d] / total;
   }
}
#endif



#ifdef USELOGLIKELIHOOD
// p(theta|x,lambda)
Real dir_sampled::getlogmu_posterior()
{
   // need to calculate it!
   // Dir(alpha+N_1, alpha+N_2, ..., alpha+N_D)
   // Gamma(alpha*D+N) / [Gamma(alpha+N_1)*Gamma(alpha+N_2)*...*Gamma(alpha+N_D)]
   // x stick_1^(alpha+N_1-1) stick_2^(alpha+N_2-1) ...
   Real tlogmu_posterior = gsl_sf_lngamma(alphah*D + N);

   #pragma omp parallel for reduction(+:tlogmu_posterior)
   for (int d=0; d<D; d++)
   {
      unordered_map<int, int>::const_iterator it = indices_counts.find(d);
      int Nd = (it!=indices_counts.end()) ? it->second : 0;
      tlogmu_posterior += (alphah+Nd-1)*param.logpi[d] - gsl_sf_lngamma(alphah + Nd);
   }
   logmu_posterior = tlogmu_posterior;
   return logmu_posterior;
}
#else
// p(theta|x,lambda)
Real dir_sampled::getlogmu_posterior()
{
   // need to calculate it!
   // Dir(alpha+N_1, alpha+N_2, ..., alpha+N_D)
   // Gamma(alpha*D+N) / [Gamma(alpha+N_1)*Gamma(alpha+N_2)*...*Gamma(alpha+N_D)]
   // x stick_1^(alpha+N_1-1) stick_2^(alpha+N_2-1) ...
   Real tlogmu_posterior = gsl_sf_lngamma(alphah*D + N);

   #pragma omp parallel for reduction(+:tlogmu_posterior)
   for (int d=0; d<D; d++)
   {
      unordered_map<int, int>::const_iterator it = indices_counts.find(d);
      int Nd = (it!=indices_counts.end()) ? it->second : 0;
      tlogmu_posterior += (alphah+Nd-1)*log(param.logpi[d]) - gsl_sf_lngamma(alphah + Nd);
   }
   logmu_posterior = tlogmu_posterior;
   return logmu_posterior;
}
#endif
Real dir_sampled::getlogmu_prior() const
{
   return logmu_prior;
}
Real dir_sampled::Jdivergence(const dir_sampled &other)
{
   // ignore the det(cov) because it will be cancelled out
   return param.Jdivergence(other.param);
}





void dir_sampled::rem_data(arr(long) indices, arr(Real) values, long nnz)
{
   // update the sufficient stats and the N
   if (N<=0) mexErrMsgTxt("Removing from empty cluster!\n");
   N--;

   if (N==0)
      empty();
   else
   {
      for (long di=0; di<nnz; di++)
      {
         // find it
         unordered_map<int,int>::iterator itr = indices_counts.find(indices[di]);
         if (itr == indices_counts.end())
         {
            mexPrintf("N=%d, d=%d, v=%d\n", N, (int)indices[di], (int)values[di]);
            mexErrMsgTxt("Removing something that didn't exist\n");
         }
         if ((itr->second -= values[di]) == 0) // subtract
            indices_counts.erase(itr); // remove if zero
      }
   }
   temp_counts_invalid = true;
}
void dir_sampled::add_data_init(arr(long) indices, arr(Real) values, long nnz)
{
   N++;

   for (long di=0; di<nnz; di++)
   {
      // try to insert
      std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = indices_counts.emplace(indices[di],values[di]);
      // see if it was found.  if it was found, just add the counts
      if (!itrFnd.second)
         itrFnd.first->second += values[di];
   }
   temp_counts_invalid = true;
}
void dir_sampled::add_data(arr(long) indices, arr(Real) values, long nnz)
{
   add_data_init(indices, values, nnz);
}




void dir_sampled::rem_data(unordered_map<int,int> &indices_counts2)
{
   for (unordered_map<int,int>::iterator it = indices_counts2.begin(); it!=indices_counts2.end(); it++)
   {
      int index = it->first;
      int count = it->second;
      N -= count;
      if ((indices_counts[index]-=count) == 0)
         indices_counts.erase(index);
   }
   temp_counts_invalid = true;
}
void dir_sampled::add_data(unordered_map<int,int> &indices_counts2)
{
   for (unordered_map<int,int>::iterator itr=indices_counts2.begin(); itr!=indices_counts2.end(); itr++)
   {
      int index = itr->first;
      int count = itr->second;
      N += count;

      // try to insert
      std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = indices_counts.emplace(index,count);
      // see if it was found.  if it was found, just add the counts
      if (!itrFnd.second)
         itrFnd.first->second += count;
   }
   temp_counts_invalid = true;
}







void dir_sampled::rem_data_dec(long index)
{
   // update the sufficient stats and the N
   if (N<=0) mexErrMsgTxt("Removing from empty cluster!\n");
   N--;

   if (N==0)
      empty();
   else
   {
      // find it
      unordered_map<int,int>::iterator itr = indices_counts.find(index);
      if (itr == indices_counts.end())
         mexErrMsgTxt("Removing something that didn't exist\n");
      if ((--(itr->second)) == 0) // subtract
         indices_counts.erase(itr); // remove if zero
   }
   temp_counts_invalid = true;
}
void dir_sampled::add_data_inc_init(long index)
{
   N++;
   std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = indices_counts.emplace(index,1);
   // see if it was found.  if it was found, just add the counts
   if (!itrFnd.second)
      itrFnd.first->second ++;
   temp_counts_invalid = true;
}
void dir_sampled::add_data_inc(long index)
{
   add_data_inc_init(index);
}


void dir_sampled::prepare_indices_counts_itr()
{
   if (temp_counts_invalid)
   {
      if (temp_counts_N != indices_counts.size())
      {
         if (temp_counts_itr!=NULL) delete[] temp_counts_itr;
         if (temp_indices_itr!=NULL) delete[] temp_indices_itr;
         temp_counts_N = indices_counts.size();
         temp_counts_itr = new int*[temp_counts_N];
         temp_indices_itr = new int[temp_counts_N];
      }

      int i = 0;
      for (unordered_map<int,int>::iterator itr=indices_counts.begin(); itr!=indices_counts.end(); itr++, i++)
      {
         temp_indices_itr[i] = itr->first;
         temp_counts_itr[i] = &(itr->second);
      }
      temp_counts_invalid = false;
   }
}

