#include "dir_sampled_full.h"

#ifndef pi
#define pi 3.14159265
#endif

#ifndef clogpi
#define clogpi 1.144729885849
#endif

// --------------------------------------------------------------------------
// -- dir_sampled_full
// --   constructor; initializes to empty
// --------------------------------------------------------------------------
dir_sampled_full::dir_sampled_full() :
   D(0), Nthreads(0), alphah(0), N(0), param(0), rArr(NULL), indices_counts(0,0)
{
   //mexErrMsgTxt("shouldn't be here\n");
}

// --------------------------------------------------------------------------
// -- dir_sampled_full
// --   copy constructor;
// --------------------------------------------------------------------------
dir_sampled_full::dir_sampled_full(const dir_sampled_full& that)
{
   copy(that);
}
// --------------------------------------------------------------------------
// -- operator=
// --   assignment operator
// --------------------------------------------------------------------------
dir_sampled_full& dir_sampled_full::operator=(const dir_sampled_full& that)
{
   if (this != &that)
   {
      cleanup();
      copy(that);
   }
   return *this;
}
// --------------------------------------------------------------------------
// -- copy
// --   returns a copy of this
// --------------------------------------------------------------------------
void dir_sampled_full::copy(const dir_sampled_full& that, bool isClean, bool copyIndices_counts)
{
   if (!isClean)
      cleanup();
   N = that.N;
   Nthreads = that.Nthreads;
   D = that.D;
   alphah = that.alphah;
   indices_counts.resize(D);

   if (copyIndices_counts)
      memcpy(indices_counts.data(), that.indices_counts.data(), sizeof(int)*D);

   param = that.param;

   rArr = allocate_memory<gsl_rng*>(Nthreads);
   for (int t=0; t<Nthreads; t++)
      rArr[t] = initialize_gsl_rand(rand());

   gammaln = that.gammaln;
}
// --------------------------------------------------------------------------
// -- copyNoNewMem
// --   copies data but doesn't create new memory
// --------------------------------------------------------------------------
void dir_sampled_full::copyNoNewMem(const dir_sampled_full& that, bool copyIndices_counts)
{
   N = that.N;
   Nthreads = that.Nthreads;
   D = that.D;
   alphah = that.alphah;
   indices_counts.resize(D);

   if (copyIndices_counts)
      memcpy(indices_counts.data(), that.indices_counts.data(), sizeof(int)*D);

   param = that.param;
   gammaln = that.gammaln;
}


// --------------------------------------------------------------------------
// -- dir_sampled_full
// --   constructor; intializes to all the values given
// --------------------------------------------------------------------------
dir_sampled_full::dir_sampled_full(int _Nthreads, int _D, Real _alphah) :
   N(0), Nthreads(_Nthreads), D(_D), alphah(_alphah), param(_D), indices_counts(_D,0)
{
   if (fabs(alphah - (Real)((int)alphah)) < 1e-30)
      gammaln = &gammalnint;
   else
      gammaln = &gsl_sf_lngamma;


   rArr = allocate_memory<gsl_rng*>(Nthreads);
   for (int t=0; t<Nthreads; t++)
      rArr[t] = initialize_gsl_rand(rand());
   clear();
}




// --------------------------------------------------------------------------
// -- ~dir_sampled_full
// --   destructor
// --------------------------------------------------------------------------
dir_sampled_full::~dir_sampled_full()
{
   cleanup();
}
// --------------------------------------------------------------------------
// -- ~cleanup
// --   deletes all the memory allocated by this
// --------------------------------------------------------------------------
void dir_sampled_full::cleanup()
{
   indices_counts.clear();
   if (rArr)
   {
      for (int t=0; t<Nthreads; t++)
         if (rArr[t]!=NULL)
            gsl_rng_free(rArr[t]);
      deallocate_memory(rArr);
   }
}

// --------------------------------------------------------------------------
// -- clear
// --   Empties out the statistics and posterior hyperparameters and the
// -- sampled parameters. This basically deletes all notions of data from
// -- the node.
// --------------------------------------------------------------------------
void dir_sampled_full::clear()
{
   empty();
   update_posteriors();
}
// --------------------------------------------------------------------------
// -- empty
// --   Empties out the statistics and posterior hyperparameters, but not
// -- the sampled parameters
// --------------------------------------------------------------------------
void dir_sampled_full::empty()
{
   N = 0;
   memset(indices_counts.data(),0,sizeof(int)*D);
}

bool dir_sampled_full::isempty() const               { return (N==0); }
int dir_sampled_full::getN() const                   { return N;}
int dir_sampled_full::getD() const                   { return D;}
arr(Real) dir_sampled_full::get_logpi() const      { return param.logpi;}
gsl_rng* dir_sampled_full::get_r()                   { return rArr[0];}

void dir_sampled_full::set_categorical(categorical &other)
{
   param = other;
}
void dir_sampled_full::set_categorical(arr(Real) _logpi)
{
   memcpy(param.logpi, _logpi, sizeof(Real)*D);
}

categorical* dir_sampled_full::get_categorical()
{
   return &param;
}



// --------------------------------------------------------------------------
// -- update_posteriors
// --   Updates the posterior hyperparameters
// --------------------------------------------------------------------------
void dir_sampled_full::update_posteriors()
{
   }

void dir_sampled_full::update_posteriors_sample()
{
   update_posteriors();
   sample();
}

// helper function to merge to indices_counts
int dir_sampled_full::merge_indices_counts(vector<int> &indices_counts2, bool clear)
{
   int numAdded = 0;
   #pragma omp parallel for reduction(+:numAdded)
   for (int d=0; d<D; d++)
   {
      indices_counts[d] += indices_counts2[d];
      numAdded += indices_counts2[d];
   }
   if (clear)
      memset(indices_counts2.data(),0,sizeof(int)*D);
   return numAdded;
}

void dir_sampled_full::temp_merge_with(dir_sampled_full* other)
{
   if (other!=NULL)
   {
      N += other->N;
      merge_indices_counts(other->indices_counts, false);
      update_posteriors_sample();
   }
}

void dir_sampled_full::merge_with(dir_sampled_full* other, bool doSample)
{
   if (other!=NULL)
      merge_with(*other, doSample);
}
void dir_sampled_full::merge_with(dir_sampled_full* other1, dir_sampled_full* other2, bool doSample)
{
   if (other1==NULL && other2!=NULL)
      merge_with(*other2,doSample);
   else if (other1!=NULL && other2==NULL)
      merge_with(*other1,doSample);
   else if (other1!=NULL && other2!=NULL)
      merge_with(*other1, *other2, doSample);
}
void dir_sampled_full::merge_with(dir_sampled_full other, bool doSample)
{
   N += other.N;
   merge_indices_counts(other.indices_counts);
   if (doSample)
      update_posteriors_sample();
   else
      update_posteriors();
}
void dir_sampled_full::merge_with(dir_sampled_full other1, dir_sampled_full other2, bool doSample)
{
   N += other1.N + other2.N;
   merge_indices_counts(other1.indices_counts);
   merge_indices_counts(other2.indices_counts);
   if (doSample)
      update_posteriors_sample();
   else
      update_posteriors();
}
void dir_sampled_full::set_stats(int _N, const int* _indices_counts)
{
   N = _N;
   memcpy(indices_counts.data(), _indices_counts, sizeof(int)*D);
   //indices_counts.clear();
   //merge_indices_counts(_indices_counts);
}
void dir_sampled_full::set_stats(int _N, const int* indices_counts1, const int* indices_counts2)
{
   N = _N;
   #pragma omp parallel for
   for (int d=0; d<D; d++)
      indices_counts[d] = indices_counts1[d] + indices_counts2[d];
}
void dir_sampled_full::set_stats(int _N, const int* indices_counts1, const int* indices_counts2,
   const int* indices_counts3)
{
   N = _N;
   #pragma omp parallel for
   for (int d=0; d<D; d++)
      indices_counts[d] = indices_counts1[d] + indices_counts2[d] + indices_counts3[d];
}




Real dir_sampled_full::data_loglikelihood() const
{
   return param.data_loglikelihood(indices_counts);
}

Real dir_sampled_full::data_loglikelihood_marginalized()
{
   Real log_likelihood = 0;
   int nnz = 0;
   #pragma omp parallel for reduction(+:log_likelihood,nnz)
   for (int d=0; d<D; d++) if (indices_counts[d]>0)
   {
      log_likelihood += gammaln(indices_counts[d] + alphah);
      nnz++;
   }

   log_likelihood += -nnz*gammaln(alphah) + gammaln(D*alphah) - gammaln(N+D*alphah);

   return log_likelihood;
}

Real dir_sampled_full::data_loglikelihood_marginalized_testmerge(const dir_sampled_full *other) const
{
   return data_loglikelihood_marginalized_testmerge(other->N, other->indices_counts.data());
}
Real dir_sampled_full::data_loglikelihood_marginalized_testmerge(const int otherN, const int* otherindices_counts) const
{
   Real log_likelihood = 0;
   int nnz = 0;
   #pragma omp parallel for reduction(+:log_likelihood,nnz)
   for (int d=0; d<D; d++)
   {
      int count = indices_counts[d] + otherindices_counts[d];
      if (count>0)
      {
         log_likelihood += gammaln(count + alphah);
         nnz++;
      }
   }

   log_likelihood += -nnz*gammaln(alphah) + gammaln(D*alphah) - gammaln(N+otherN+D*alphah);
   return log_likelihood;
}

#ifdef USELOGLIKELIHOOD
void dir_sampled_full::sample()
{
   Real total = 0;
   arr(Real) logpi = param.logpi;
   // first sample all assuming zero counts
   #pragma omp parallel for reduction(+:total)
   for (int d=0; d<D; d++)
   {
      logpi[d] = gsl_ran_gamma(rArr[omp_get_thread_num()], alphah+indices_counts[d], 1);
      total += logpi[d];
   }
   total = log(total);
   #pragma omp parallel for// reduction(+:logmu_prior,logmu_posterior)
   for (int d=0; d<D; d++)
   {
      logpi[d] = log(logpi[d]) - total;
   }
}
#else
void dir_sampled_full::sample()
{
   Real total = 0;
   arr(Real) logpi = param.logpi;
   // first sample all assuming zero counts
   #pragma omp parallel for reduction(+:total)
   for (int d=0; d<D; d++)
   {
      logpi[d] = gsl_ran_gamma(rArr[omp_get_thread_num()], alphah+indices_counts[d], 1);
      total += logpi[d];
   }

   #pragma omp parallel for// reduction(+:logmu_prior,logmu_posterior)
   for (int d=0; d<D; d++)
   {
      logpi[d] = logpi[d] / total;
   }
}
#endif



#ifdef USELOGLIKELIHOOD
// p(theta|x,lambda)
Real dir_sampled_full::getlogmu_posterior()
{
   // need to calculate it!
   // Dir(alpha+N_1, alpha+N_2, ..., alpha+N_D)
   // Gamma(alpha*D+N) / [Gamma(alpha+N_1)*Gamma(alpha+N_2)*...*Gamma(alpha+N_D)]
   // x stick_1^(alpha+N_1-1) stick_2^(alpha+N_2-1) ...
   Real tlogmu_posterior = gsl_sf_lngamma(alphah*D + N);

   #pragma omp parallel for reduction(+:tlogmu_posterior)
   for (int d=0; d<D; d++)
   {
      int Nd = indices_counts[d];
      tlogmu_posterior += (alphah+Nd-1)*param.logpi[d] - gsl_sf_lngamma(alphah + Nd);
   }
   logmu_posterior = tlogmu_posterior;
   return logmu_posterior;
}
#else
// p(theta|x,lambda)
Real dir_sampled_full::getlogmu_posterior()
{
   // need to calculate it!
   // Dir(alpha+N_1, alpha+N_2, ..., alpha+N_D)
   // Gamma(alpha*D+N) / [Gamma(alpha+N_1)*Gamma(alpha+N_2)*...*Gamma(alpha+N_D)]
   // x stick_1^(alpha+N_1-1) stick_2^(alpha+N_2-1) ...
   Real tlogmu_posterior = gsl_sf_lngamma(alphah*D + N);

   #pragma omp parallel for reduction(+:tlogmu_posterior)
   for (int d=0; d<D; d++)
   {
      int Nd = indices_counts[d];
      tlogmu_posterior += (alphah+Nd-1)*log(param.logpi[d]) - gsl_sf_lngamma(alphah + Nd);
   }
   logmu_posterior = tlogmu_posterior;
   return logmu_posterior;
}
#endif
Real dir_sampled_full::getlogmu_prior() const
{
   return logmu_prior;
}
Real dir_sampled_full::Jdivergence(const dir_sampled_full &other)
{
   // ignore the det(cov) because it will be cancelled out
   return param.Jdivergence(other.param);
}



void dir_sampled_full::rem_data(vector<int> &indices_counts2)
{
   rem_data(indices_counts2.data());
}
void dir_sampled_full::add_data(vector<int> &indices_counts2)
{
   add_data(indices_counts2.data());
}
void dir_sampled_full::rem_data(int* indices_counts2)
{
   int tN = 0;
   #pragma omp parallel for reduction(+:tN)
   for (int d=0; d<D; d++)
   {
      indices_counts[d] -= indices_counts2[d];
      N += -indices_counts2[d];
   }
   N -= tN;
}
void dir_sampled_full::add_data(int* indices_counts2)
{
   int tN = 0;
   #pragma omp parallel for reduction(+:tN)
   for (int d=0; d<D; d++)
   {
      indices_counts[d] += indices_counts2[d];
      tN += indices_counts2[d];
   }
   N += tN;
}







void dir_sampled_full::rem_data_dec(long index)
{
   // update the sufficient stats and the N
   if (N<=0) mexErrMsgTxt("Removing from empty cluster!\n");
   N--;

   if (N==0)
      empty();
   else
   {
      indices_counts[index]--;
   }
}
void dir_sampled_full::add_data_inc_init(long index)
{
   N++;
   indices_counts[index]++;
}
void dir_sampled_full::add_data_inc(long index)
{
   add_data_inc_init(index);
}
