#ifndef _CATEGORICAL_H_INCLUDED_
#define _CATEGORICAL_H_INCLUDED_

#include "matrix.h"
#include "mex.h"
#include <math.h>
#include "array.h"
#include "gsl/gsl_sf_gamma.h"

#include "helperMEX.h"
#include "debugMEX.h"

#include <boost/unordered_map.hpp>
using boost::unordered_map;
#include <vector>
using std::vector;

typedef float Real;

class categorical
{
   // instantiated gaussian parameters
   int D;
   arr(Real) logpi;

public:
   // --------------------------------------------------------------------------
   // -- categorical
   // --   constructor; initializes to empty
   // --------------------------------------------------------------------------
   categorical();
   // --------------------------------------------------------------------------
   // -- categorical
   // --   copy constructor;
   // --------------------------------------------------------------------------
   categorical(const categorical& that);
   // --------------------------------------------------------------------------
   // -- operator=
   // --   assignment operator
   // --------------------------------------------------------------------------
   categorical& operator=(const categorical& that);
   // --------------------------------------------------------------------------
   // -- copy
   // --   returns a copy of this
   // --------------------------------------------------------------------------
   void copy(const categorical& that);
   // --------------------------------------------------------------------------
   // -- categorical
   // --   constructor; intializes to all the values given
   // --------------------------------------------------------------------------
   categorical(int _D);

   // --------------------------------------------------------------------------
   // -- ~categorical
   // --   destructor
   // --------------------------------------------------------------------------
   virtual ~categorical();

   // --------------------------------------------------------------------------
   // -- ~cleanup
   // --   deletes all the memory allocated by this
   // --------------------------------------------------------------------------
   virtual void cleanup();

   void set_values(arr(Real) _logpi);

#ifdef USELOGLIKELIHOOD
   Real predictive_loglikelihood(long index) const;
   Real predictive_loglikelihood(unordered_map<int,int> &indices_counts) const;
   Real predictive_loglikelihood(arr(long) indices, arr(Real) values, long nnz) const;
#else
   Real predictive_likelihood(long index) const;
   Real predictive_likelihood(unordered_map<int,int> &indices_counts) const;
   Real predictive_likelihood(arr(long) indices, arr(Real) values, long nnz) const;
#endif
   Real data_loglikelihood(const unordered_map<int,int> &indices_counts) const;
   Real data_loglikelihood(const vector<int> &indices_counts) const;
   Real Jdivergence(const categorical &other);

   friend class dir_sampled;
   friend class dir_sampled_full;
   friend class cluster_sampled;
};

inline void categorical::set_values(arr(Real) _logpi)
{
   memcpy(logpi, _logpi, sizeof(Real)*D);
}


#ifdef USELOGLIKELIHOOD
inline Real categorical::predictive_loglikelihood(long index) const
{
   return logpi[index];
}
inline Real categorical::predictive_loglikelihood(unordered_map<int,int> &indices_counts) const
{
   Real loglikelihood = 0;
   for (unordered_map<int,int>::iterator it = indices_counts.begin(); it!=indices_counts.end(); it++)
   {
      long d = it->first;
      int count = it->second;
      loglikelihood += count*logpi[d];
   }
   return loglikelihood;
}
inline Real categorical::predictive_loglikelihood(arr(long) indices, arr(Real) values, long nnz) const
{
   Real loglikelihood = 0;
   for (long i=0; i<nnz; i++)
   {
      long d = indices[i];
      int count = values[i];
      loglikelihood += count*logpi[d];
   }
   return loglikelihood;
}
#else
inline Real categorical::predictive_likelihood(long index) const
{
   return logpi[index];
}
inline Real categorical::predictive_likelihood(unordered_map<int,int> &indices_counts) const
{
   Real likelihood = 0;
   for (unordered_map<int,int>::iterator it = indices_counts.begin(); it!=indices_counts.end(); it++)
   {
      long d = it->first;
      int count = it->second;
      likelihood = pow(logpi[d],count);
   }
   return likelihood;
}
inline Real categorical::predictive_likelihood(arr(long) indices, arr(Real) values, long nnz) const
{
   Real likelihood = 0;
   for (long i=0; i<nnz; i++)
   {
      long d = indices[i];
      int count = values[i];
      likelihood += pow(logpi[d],count);
   }
   return likelihood;
}
#endif


inline Real categorical::data_loglikelihood(const unordered_map<int,int> &indices_counts) const
{
   int N = 0;
   Real loglikelihood = 0;
   for (unordered_map<int,int>::const_iterator itr=indices_counts.begin(); itr!=indices_counts.end(); itr++)
   {
      int d = itr->first;
      int count = itr->second;
      loglikelihood += count*logpi[d];
      N += count;
   }
   return loglikelihood;
}

inline Real categorical::data_loglikelihood(const vector<int> &indices_counts) const
{
   int N = 0;
   Real loglikelihood = 0;
   for (int d=0; d<D; d++)
   {
      int count = indices_counts[d];
      loglikelihood += count*logpi[d];
      N += count;
   }
   return loglikelihood;
}


inline Real categorical::Jdivergence(const categorical &other)
{
   Real J = 0;
   #pragma omp parallel for reduction(+:J)
   for (int d=0; d<D; d++)
   {
      Real logpi1 = logpi[d];
      Real logpi2 = other.logpi[d];
      J += (exp(logpi1)-exp(logpi2)) * (logpi1-logpi2);
      //Real temp = logpi[d] - other.logpi[d];
      //J += temp*exp(temp);
   }
   return J;
}

#endif
