// =============================================================================
// == reduction_hash2.h
// == --------------------------------------------------------------------------
// == A reduction hash class that can be used with OpenMP. This allows one to
// == perform reduction into multiple hash tables. Current OpenMP software only
// == allows the reduction of an array into a scalar. This class allows one to
// == reduce an array into one of many hash tables.
// ==
// == numThreads x K hash tables are created.
// ==
// == Notation is as follows:
// ==   numThreads - The number of OpenMP threads that will be running
// ==   K - The number of unique hash tables to reduce into.
// ==
// == --------------------------------------------------------------------------
// == Written by Jason Chang 04-14-2013 - jchang7@csail.mit.edu
// =============================================================================

#ifndef _REDUCTION_HASH2
#define _REDUCTION_HASH2

#include <string.h>
#include <cstdlib>
#include <stdlib.h>
#include <iostream>
#include <algorithm>
#include "array.h"
#include "helperMEX.h"
#include <boost/unordered_map.hpp>

using boost::unordered_map;

template <typename T>
class reduction_hash2
{
private:
   unordered_map< int, T >* data;
   int numThreads;
   int K;

public:
   // --------------------------------------------------------------------------
   // -- reduction_hash2
   // --   constructors; initializes the reduction array with a number of
   // -- threads, each with a D dimensional vector. The third parameter can be
   // -- specified to give the initializing value.
   // --------------------------------------------------------------------------
   reduction_hash2();
   reduction_hash2(int thenumThreads, int theK);
   virtual ~reduction_hash2();

   // --------------------------------------------------------------------------
   // -- clear_values
   // --   Clears the arrays
   // --------------------------------------------------------------------------
   void clear_values();

   // --------------------------------------------------------------------------
   // -- reduce_XXX
   // --   Performs the reduction "XXX" on the d^th dimension with value
   // --------------------------------------------------------------------------
   void reduce_inc(int t, int k, long d);
   void reduce_inc(int t, int k, long* ds, int nnz);
   template <typename T2>
   void reduce_add(int t, int k, long d, T2 value);
   template <typename T2>
   void reduce_add(int t, int k, long* ds, T2* values, int nnz);

   // --------------------------------------------------------------------------
   // -- final_reduce_XXX
   // --   Performs the reduction "XXX" on the threads and returns result.
   // --------------------------------------------------------------------------
   unordered_map<int,T>* final_reduce_add();
};

#endif

// --------------------------------------------------------------------------
// -- reduction_sparray
// --   constructors; initializes the reduction array with a number of
// -- threads, each with a D dimensional vector. The third parameter can be
// -- specified to give the initializing value.
// --------------------------------------------------------------------------
template <typename T>
reduction_hash2<T>::reduction_hash2() :
   numThreads(0), data(NULL)
{
}
template <typename T>
reduction_hash2<T>::reduction_hash2(int thenumThreads, int theK) :
   numThreads(thenumThreads), K(theK)
{
   data = allocate_memory< unordered_map< int, T > >(numThreads*K);

}
template <typename T>
reduction_hash2<T>::~reduction_hash2()
{
   if (data!=NULL) deallocate_memory(data);
}


// --------------------------------------------------------------------------
// -- clear_values
// --   Clears the arrays
// --------------------------------------------------------------------------
template <typename T>
void reduction_hash2<T>::clear_values()
{
   for (int t=0; t<numThreads*K; t++)
      data[t].clear();
}


// --------------------------------------------------------------------------
// -- reduce_XXX
// --   Performs the reduction "XXX" on the d^th dimension with value
// --------------------------------------------------------------------------
template <typename T>
inline void reduction_hash2<T>::reduce_inc(int t, int k, long d)
{
   std::pair< typename unordered_map<int, T>::iterator, bool> itrFnd = data[t*K+k].emplace(d,1);
   if (!itrFnd.second)
      itrFnd.first->second++;
}
template <typename T>
inline void reduction_hash2<T>::reduce_inc(int t, int k, long* ds, int nnz)
{
   int bin = t*K+k;
   for (int i=0; i<nnz; i++)
   {
      std::pair< typename unordered_map<int, T>::iterator, bool> itrFnd = data[bin].emplace(ds[i],1);
      if (!itrFnd.second)
         itrFnd.first->second++;
   }
}
template <typename T> template <typename T2>
inline void reduction_hash2<T>::reduce_add(int t, int k, long d, T2 value)
{
   std::pair< typename unordered_map<int, T>::iterator, bool> itrFnd = data[t*K+k].emplace(d,value);
   if (!itrFnd.second)
      itrFnd.first->second += value;
}
template <typename T> template <typename T2>
inline void reduction_hash2<T>::reduce_add(int t, int k, long* ds, T2* values, int nnz)
{
   int bin = t*K+k;
   for (int i=0; i<nnz; i++)
   {
      std::pair< typename unordered_map<int, T>::iterator, bool> itrFnd = data[bin].emplace(ds[i], values[i]);
      if (!itrFnd.second)
         itrFnd.first->second += values[i];
   }
}


// --------------------------------------------------------------------------
// -- final_reduce_XXX
// --   Performs the reduction "XXX" on the threads and returns result.
// --------------------------------------------------------------------------
template <typename T>
inline unordered_map<int,T>* reduction_hash2<T>::final_reduce_add()
{
   for (int t=1; t<numThreads; t++)
      for (int k=0; k<K; k++)
      {
         for (unordered_map<int,int>::iterator itr=data[t*K+k].begin(); itr!=data[t*K+k].end(); itr++)
         {
            int index = itr->first;
            int count = itr->second;

            // try to insert
            std::pair< typename unordered_map<int,int>::iterator, bool> itrFnd = data[k].emplace(index,count);
            // see if it was found.  if it was found, just add the counts
            if (!itrFnd.second)
               itrFnd.first->second += count;
         }
      }
   return data;
}



